# Copyright (c) 2018 Stephen Bunn <stephen@bunn.io>
# MIT License <https://opensource.org/licenses/MIT>

__name__ = "file-config"
__repo__ = "https://github.com/stephen-bunn/file-config"
__version__ = "0.0.7"
__description__ = "An attr's based configuration file abstraction"
__author__ = "Stephen Bunn"
__contact__ = "stephen@bunn.io"
__license__ = "MIT License"



# Copyright (c) 2018 Stephen Bunn <stephen@bunn.io>
# MIT License <https://opensource.org/licenses/MIT>

from ._common import BaseHandler


class JSONHandler(BaseHandler):
    """ The JSON serialization handler.
    """

    name = "json"
    packages = ("ujson", "json")
    options = {"indent": 0}

    def on_json_dumps(self, json, dictionary, **kwargs):
        """ The :mod:`json` dumps method.

        :param module json: The ``json`` module
        :param dict dictionary: The dictionary instance to serialize
        :param int indent: The amount of spaces to use for indentation,
            defaults to 0, optional
        :returns: The json serialization of the given ``dictionary``
        :rtype: str
        """

        return json.dumps(dictionary, indent=kwargs.get("indent", 0))

    def on_json_loads(self, json, content):
        """ The :mod:`json` loads method.

        :param module json: The ``json`` module
        :param str content: The content to deserialize
        :returns: The deserialized dictionary
        :rtype: dict
        """

        return json.loads(content)

    def on_ujson_dumps(self, ujson, dictionary, **kwargs):
        """ The `ujson <https://pypi.org/project/ujson/>`_ dumps method.

        :param module ujson: The ``ujson`` module
        :param dict dictionary: The dictionary instance to serialize
        :param int indent: The amount of spaces to use for indentation,
            defaults to 0, optional
        :returns: The json serialization of the given ``dictionary``
        :rtype: str
        """

        return ujson.dumps(dictionary, indent=kwargs.get("indent", 0))

    def on_ujson_loads(self, ujson, content):
        """ The `ujson <https://pypi.org/project/ujson/>`_ loads method.

        :param module ujson: The ``ujson`` module
        :param str content: The content to deserialize
        :returns: The deserialized dictionary
        :rtype: dict
        """

        return ujson.loads(content)



# Copyright (c) 2018 Stephen Bunn <stephen@bunn.io>
# MIT License <https://opensource.org/licenses/MIT>

import re
import typing
import collections

import attr

from .constants import CONFIG_KEY, REGEX_TYPE_NAME

COMPILED_PATTERN_TYPE = type(re.compile(""))


def is_config_var(var):
    return (
        isinstance(var, (attr._make.Attribute, attr._make._CountingAttr))
        and hasattr(var, "metadata")
        and CONFIG_KEY in var.metadata
    )


def is_config_type(type_):
    return (
        isinstance(type_, type)
        and hasattr(type_, "__attrs_attrs__")
        and hasattr(type_, CONFIG_KEY)
    )


def is_config(config_instance):
    return isinstance(config_instance, object) and is_config_type(type(config_instance))


def is_compiled_pattern(compiled_pattern):
    return isinstance(compiled_pattern, COMPILED_PATTERN_TYPE)


def is_builtin_type(type_):
    return isinstance(type_, type) and type_ in (
        bool,
        str,
        int,
        float,
        complex,
        list,
        tuple,
        bytes,
        bytearray,
        set,
        frozenset,
        dict,
    )


def is_typing_type(type_):
    return isinstance(type_, type) and getattr(type_, "__module__", None) == "typing"


def is_collections_type(type_):
    return (
        isinstance(type_, type) and getattr(type_, "__module__", None) == "collections"
    )


def is_regex_type(type_):
    return (
        callable(type_)
        and type_.__name__ == REGEX_TYPE_NAME
        and hasattr(type_, "__supertype__")
        and is_compiled_pattern(type_.__supertype__)
    )


def is_union_type(type_):
    return isinstance(type_, typing._Union)


def is_null_type(type_):
    if type_ in (type(None),):
        return True
    return False


def is_bool_type(type_):
    if is_builtin_type(type_):
        return type_ in (bool,)
    return False


def is_string_type(type_):
    if is_builtin_type(type_):
        return type_ in (str,)
    elif is_typing_type(type_):
        return type_.__origin__ in (typing.Text, typing.AnyStr)
    elif is_collections_type(type_):
        return type_ in (collections.UserString,)
    elif is_regex_type(type_):
        return True
    return False


def is_integer_type(type_):
    if is_builtin_type(type_):
        return type_ in (int,)
    return False


def is_number_type(type_):
    if is_builtin_type(type_):
        return type_ in (float, complex)
    return False


def is_array_type(type_):
    if is_builtin_type(type_):
        return type_ in (list, tuple, set, frozenset)
    elif is_typing_type(type_):
        return type_.__origin__ in (
            typing.List,
            typing.Tuple,
            typing.Set,
            typing.FrozenSet,
        )
    elif is_collections_type(type_):
        return type_ in (collections.deque, collections.UserList)
    return False


def is_object_type(type_):
    if is_builtin_type(type_):
        return type_ in (dict,)
    elif is_typing_type(type_):
        return type_.__origin__ in (typing.Dict,)
    elif is_collections_type(type_):
        return type_ in (
            collections.ChainMap,
            collections.Counter,
            collections.OrderedDict,
            collections.defaultdict,
            collections.UserDict,
        )
    return False


def typecast(type_, value):
    # NOTE: does not do any special validation of types before casting
    # will just raise errors on type casting failures
    if is_builtin_type(type_) or is_collections_type(type_):
        return type_(value)
    elif is_regex_type(type_):
        return typecast(str, value)
    elif is_typing_type(type_):
        base_type = type_.__extra__
        arg_types = type_.__args__

        if is_array_type(type_):
            if len(arg_types) == 1:
                item_type = arg_types[0]
                return base_type([typecast(item_type, item) for item in value])
            else:
                return base_type(value)
        elif is_object_type(type_):
            if len(arg_types) == 2:
                (key_type, item_type) = arg_types
                return base_type(
                    {
                        typecast(key_type, key): typecast(item_type, item)
                        for (key, item) in value.items()
                    }
                )
            else:
                return base_type(value)
        else:
            return base_type(value)
    else:
        return value



# Copyright (c) 2018 Stephen Bunn <stephen@bunn.io>
# MIT License <https://opensource.org/licenses/MIT>

from collections import OrderedDict

from ._common import BaseHandler


class YAMLHandler(BaseHandler):
    """ The YAML serialization handler.
    """

    name = "yaml"
    packages = ("yaml",)
    options = {}

    def on_yaml_imported(self, yaml):
        """ The `pyyaml <https://pypi.org/project/pyyaml/>`_ import hook.

        :param module yaml: The ``yaml`` module
        """

        def represent_ordereddict(dumper, data):
            """ A custom data representer for ``OrderedDict`` instances.

            .. note:: Credit to https://stackoverflow.com/a/16782282/7828560.

            :param object dumper: The dumper object
            :param collections.OrderedDict data: The ``OrderedDict`` instance
            :return: The yaml mapping node
            :rtype: yaml.MappingNode
            """

            values = []
            for (key, value) in data.items():
                values.append(
                    (dumper.represent_data(key), dumper.represent_data(value))
                )
            return yaml.MappingNode("tag:yaml.org,2002:map", values)

        yaml.add_representer(OrderedDict, represent_ordereddict)

    def on_yaml_dumps(self, yaml, dictionary):
        """ The `pyyaml <https://pypi.org/project/pyyaml/>`_ dumps method.

        :param module yaml: The ``yaml`` module
        :param dict dictionary: The dictionary to seralize
        :returns: The serialized content
        :rtype: str
        """

        return yaml.dump(dictionary)

    def on_yaml_loads(self, yaml, content):
        """ The `pyyaml <https://pypi.org/project/pyyaml/>`_ loads method.

        :param module yaml: The ``yaml`` module
        :param str content: The content to deserialize
        :returns: The deserialized dictionary
        :rtype: dict
        """

        return yaml.load(content)



# Copyright (c) 2018 Stephen Bunn <stephen@bunn.io>
# MIT License <https://opensource.org/licenses/MIT>

import abc
import sys
import warnings
import importlib


class BaseHandler(abc.ABC):
    """ The base handler that all handlers should inherit from.
    """

    @property
    def imported(self):
        """ The imported handler module.

        :return: The imported handler module.
        :rtype: module
        """

        if not hasattr(self, "_imported"):
            self._imported = self._discover_import()
        return self._imported

    @property
    def handler(self):
        if not hasattr(self, "_handler"):
            self._handler = sys.modules[self.imported]
        return self._handler

    @abc.abstractproperty
    def name(self):
        """ A unique name for the handler.

        .. note:: Handler names are used as the suffix of the dynamically set methods
            for all config instances.

        :raises NotImplementedError: Must be implemented by subclasses
        """

        raise NotImplementedError(f"subclasses must implement 'name'")

    @abc.abstractproperty
    def packages(self):
        """ A tuple of supported packages to use as handler modules.

        :raises NotImplementedError: Must be implemented by subclasses
        """

        raise NotImplementedError(f"subclasses must implement 'packages'")

    @abc.abstractproperty
    def options(self):
        """ A dictionary of dumping options that should be supported by the handler.

        :raises NotImplementedError: Must be implemented by subclasses
        """

        raise NotImplementedError(f"subclasses must implement 'options'")

    @classmethod
    def available(self):
        """ True if any of the supported modules from ``packages`` is available for use.

        :return: True if any modules from ``packages`` exist
        :rtype: bool
        """

        for module_name in self.packages:
            if importlib.util.find_spec(module_name):
                return True
        return False

    def _discover_import(self):
        """ Discovers and imports the best available module from ``packages``.

        :raises ModuleNotFoundError: If no module is available
        :return: The name of the module to use
        :rtype: str
        """

        for module_name in self.packages:
            spec = importlib.util.find_spec(module_name)
            if spec is not None:
                importlib.import_module(module_name)
                imported_hook = getattr(self, f"on_{module_name}_imported", None)
                if callable(imported_hook):
                    imported_hook(sys.modules[module_name])
                return module_name
        raise ModuleNotFoundError(f"no modules in {self.packages!r} found")

    def _prefer_package(self, package):
        if package not in self.packages:
            raise ValueError(
                f"prefered package {package!r} does not exist, "
                f"allowed are {self.packages!r}"
            )
        self.packages = (package,)

    def dumps(self, instance, prefer=None, **kwargs):
        """ An abstract dumps method which dumps an instance into the subclasses format.

        :param object instance: The instance to dump
        :raises ValueError: If dump handler does not provide handler method
        :return: The dumped content
        :rtype: str
        """

        if isinstance(prefer, str):
            self._prefer_package(prefer)

        dumps_hook_name = f"on_{self.imported}_dumps"
        dumps_hook = getattr(self, dumps_hook_name, None)
        if not callable(dumps_hook):
            raise ValueError(
                f"no dumps handler for {self.imported!r}, requires method "
                f"{dumps_hook_name!r} in {self!r}"
            )

        extras = self.options.copy()
        for (key, value) in kwargs.items():
            if key not in extras.keys():
                warnings.warn(
                    f"handler 'dumps_{self.name!s}' does not support {key!r} argument"
                )
            else:
                extras[key] = value
        return dumps_hook(self.handler, instance, **extras)

    def loads(self, content, prefer=None):
        """ An abstract loads method which loads an instance from some content.

        :param str content: The content to load from
        :raises ValueError: If load handler does not provided handler method
        :return: A dictionary converted from the given content
        :rtype: dict
        """

        if isinstance(prefer, str):
            self._prefer_package(prefer)

        loads_hook_name = f"on_{self.imported}_loads"
        loads_hook = getattr(self, loads_hook_name, None)
        if not callable(loads_hook):
            raise ValueError(
                f"no loads handler for {self.imported!r}, requires method "
                f"{loads_hook_name!r} in {self!r}"
            )
        return loads_hook(self.handler, content)

    def dump(self, instance, file_object, prefer=None):
        """ An abstract method that dumps to a given file object.

        :param object instance: The instance to dump
        :param file file_object: The file object to dump to
        """

        file_object.write(self.dumps(instance, prefer=prefer))

    def load(self, file_object, prefer=None):
        """ An abstract method that loads from a given file object.

        :param file file_object: The file object to load from
        :returns: A dictionary converted from the content of the given file object
        :rtype: dict
        """

        return self.loads(file_object.read(), prefer=prefer)



# Copyright (c) 2018 Stephen Bunn <stephen@bunn.io>
# MIT License <https://opensource.org/licenses/MIT>

from ._common import BaseHandler


class MsgpackHandler(BaseHandler):
    """ The Message Pack serialization handler.
    """

    name = "msgpack"
    packages = ("msgpack",)
    options = {}

    def on_msgpack_dumps(self, msgpack, dictionary):
        """ The `msgpack <https://pypi.org/project/msgpack/>`_ dumps method.

        :param module msgpack: The ``msgpack`` module
        :param dict dictionary: The dictionary instance to serialize
        :returns: The serialized content
        :rtype: str
        """

        return msgpack.dumps(dictionary)

    def on_msgpack_loads(self, msgpack, content):
        """ The `msgpack <https://pypi.org/project/msgpack/>`_ loads method.

        :param module msgpack: The ``msgpack`` module
        :param dict content: The serialized content to deserialize
        :returns: The deserialized dictionary
        :rtype: dict
        """

        return msgpack.loads(content, raw=False)



# Copyright (c) 2018 Stephen Bunn <stephen@bunn.io>
# MIT License <https://opensource.org/licenses/MIT>

import typing
from functools import partialmethod
from collections import OrderedDict

import attr
import jsonschema

from . import handlers
from .utils import (
    typecast,
    is_config,
    is_array_type,
    is_config_var,
    is_config_type,
    is_object_type,
    is_string_type,
    is_typing_type,
)
from .constants import CONFIG_KEY
from .schema_builder import build_schema


@attr.s(slots=True)
class _ConfigEntry(object):
    """ Configuration entry.
    """

    type = attr.ib(default=None)
    default = attr.ib(type=str, default=None)
    name = attr.ib(type=str, default=None)
    title = attr.ib(type=str, default=None)
    description = attr.ib(type=str, default=None)
    required = attr.ib(type=bool, default=True)
    examples = attr.ib(type=list, default=None)
    min = attr.ib(type=int, default=None)
    max = attr.ib(type=int, default=None)
    unique = attr.ib(type=bool, default=None)
    contains = attr.ib(type=list, default=None)


def _handle_dumps(self, handler, **kwargs):
    """ Dumps caller, used by partial method for dynamic handler assignments.

    :param object handler: The dump handler
    :return: The dumped string
    :rtype: str
    """

    return handler.dumps(to_dict(self), **kwargs)


def _handle_dump(self, handler, file_object, **kwargs):
    """ Dump caller, used by partial method for dynamic handler assignments.

    :param object handler: The dump handler
    :param file file_object: The file object to dump to
    :return: The dumped string
    :rtype: str
    """

    return handler.dump(to_dict(self), file_object, **kwargs)


@classmethod
def _handle_loads(cls, handler, content):
    """ Loads caller, used by partial method for dynamic handler assignments.

    :param object handler: The loads handler
    :param str content: The content to load from
    :return: The loaded instance
    :rtype: object
    """

    return from_dict(cls, handler.loads(content))


@classmethod
def _handle_load(cls, handler, file_object):
    """ Loads caller, used by partial method for dynamic handler assignments.

    :param object handler: The loads handler
    :param file file_object: The file object to load from
    :return: The loaded instance
    :rtype: object
    """

    return from_dict(cls, handler.load(file_object))


def config(maybe_cls=None, these=None, title=None, description=None):
    """ File config class decorator.

    :param class maybe_cls: The class to inherit from, defaults to None, optional
    :param dict these: A dictionary of str to ``file_config.var`` to use as attribs
    :param str title: The title of the config, defaults to None, optional
    :param str description: A description of the config, defaults to None, optional
    :return: Config wrapped class
    :rtype: class
    """

    def wrap(config_cls):
        """ The wrapper function.

        :param class config_cls: The class to wrap
        :return: The config_cls wrapper
        :rtype: class
        """

        setattr(config_cls, CONFIG_KEY, dict(title=title, description=description))
        # dynamically assign available handlers to the wrapped class
        for handler_name in handlers.__all__:
            handler = getattr(handlers, handler_name)
            if handler.available:
                handler = handler()
                setattr(
                    config_cls,
                    f"dumps_{handler.name}",
                    partialmethod(_handle_dumps, handler),
                )
                setattr(
                    config_cls,
                    f"dump_{handler.name}",
                    partialmethod(_handle_dump, handler),
                )
                setattr(
                    config_cls,
                    f"loads_{handler.name}",
                    partialmethod(_handle_loads, handler),
                )
                setattr(
                    config_cls,
                    f"load_{handler.name}",
                    partialmethod(_handle_load, handler),
                )
        config_vars = these if isinstance(these, dict) else None
        return attr.s(config_cls, these=config_vars, slots=True)

    if maybe_cls is None:
        return wrap
    else:
        return wrap(maybe_cls)


def var(
    type=None,
    default=None,
    name=None,
    title=None,
    description=None,
    required=True,
    examples=None,
    min=None,
    max=None,
    unique=None,
    contains=None,
    **kwargs,
):
    """ Creates a config variable.

    :param type type: The expected type of the variable, defaults to None, optional
    :param default: The default value of the var, defaults to None, optional
    :param str name: The serialized name of the variable, defaults to None, optional
    :param str title: The validation title of the variable, defaults to None, optional
    :param str description: The validation description of the variable,
        defaults to None, optional
    :param bool required: Flag to indicate if variable is required during validation,
        defaults to True, optional
    :param list examples: A list of validation examples, if necessary,
        defaults to None, optional
    :param int min: The minimum constraint of the variable, defaults to None, optional
    :param int max: The maximum constraint of the variable, defaults to None, optional
    :param bool unique: Flag to indicate if variable should be unique,
        may not apply to all variable types, defaults to None, optional
    :param contains: Value that list varaible should contain in validation,
        may not apply to all variable types, defaults to None, optional
    :return: A new config variable
    :rtype: attr.Attribute
    """

    kwargs.update(dict(default=default, type=type))
    return attr.ib(
        metadata={
            CONFIG_KEY: _ConfigEntry(
                type=type,
                default=default,
                name=name,
                title=title,
                description=description,
                required=required,
                examples=examples,
                min=min,
                max=max,
                unique=unique,
                contains=contains,
            )
        },
        **kwargs,
    )


def make_config(name, var_dict, title=None, description=None, **kwargs):
    """ Creates a config instance from scratch.

    :param str name: The name of the config
    :param dict var_dict: The dictionary of config variable definitions
    :param str title: The title of the config, defaults to None, optional
    :param str description: The description of the config, defaults to None, optional
    :return: A new config class
    :rtype: class
    """

    return config(
        attr.make_class(name, attrs={}, **kwargs),
        these=var_dict,
        title=title,
        description=description,
    )


def _build(config_cls, dictionary):
    """ Builds an instance of ``config_cls`` using ``dictionary``.

    :param type config_cls: The class to use for building
    :param dict dictionary: The dictionary to use for building ``config_cls``
    :return: An instance of ``config_cls``
    :rtype: object
    """

    if not is_config_type(config_cls):
        raise ValueError(
            f"cannot build {config_cls!r} from {dictionary!r}, "
            f"{config_cls!r} is not a config"
        )

    # perform jsonschema validation on the given dictionary
    # (simplifys dynamic typecasting)
    jsonschema.validate(dictionary, build_schema(config_cls))

    kwargs = {}
    for var in attr.fields(config_cls):
        if not is_config_var(var):
            continue

        entry = var.metadata[CONFIG_KEY]
        arg_key = entry.name if entry.name else var.name
        arg_default = var.default if var.default else None

        if is_array_type(entry.type):
            if is_typing_type(entry.type) and len(entry.type.__args__) > 0:
                nested_type = entry.type.__args__[0]
                if is_config_type(nested_type):
                    kwargs[var.name] = [
                        _build(nested_type, item)
                        for item in dictionary.get(arg_key, [])
                    ]
                else:
                    kwargs[var.name] = typecast(entry.type, dictionary.get(arg_key, []))
        elif is_object_type(entry.type):
            item = dictionary.get(arg_key, {})
            if is_typing_type(entry.type) and len(entry.type.__args__) == 2:
                (key_type, value_type) = entry.type.__args__
                kwargs[var.name] = {
                    key: _build(value_type, value)
                    if is_config_type(value_type)
                    else typecast(value_type, value)
                    for (key, value) in item.items()
                }
            else:
                kwargs[var.name] = typecast(entry.type, item)
        elif is_config_type(entry.type):
            kwargs[var.name] = _build(entry.type, dictionary.get(arg_key, arg_default))
        else:
            # TODO: handle correct type casting logic based on builtin or typing types
            kwargs[var.name] = typecast(
                entry.type, dictionary.get(arg_key, arg_default)
            )

    return config_cls(**kwargs)


def _dump(config_instance, dict_type=OrderedDict):
    """ Dumps an instance from ``instance`` to a dictionary type mapping.

    :param object instance: The instance to serialized to a dictionary
    :param object dict_type: Some dictionary type, defaults to ``OrderedDict``
    :return: Dumped dictionary
    :rtype: collections.OrderedDict (or instance of ``dict_type``)
    """

    if not is_config(config_instance):
        raise ValueError(
            f"cannot dump instance {config_instance!r} to dict, "
            "instance is not a config class"
        )

    result = dict_type()
    for var in attr.fields(config_instance.__class__):
        if not is_config_var(var):
            continue

        entry = var.metadata[CONFIG_KEY]
        dump_key = entry.name if entry.name else var.name
        dump_default = var.default if var.default else None

        if is_array_type(entry.type):
            items = getattr(config_instance, var.name, [])
            if items is not None:
                result[dump_key] = [
                    (_dump(item, dict_type=dict_type) if is_config(item) else item)
                    for item in items
                ]
        else:
            if is_config_type(entry.type):
                result[dump_key] = _dump(
                    getattr(config_instance, var.name, {}), dict_type=dict_type
                )
            else:
                dump_value = getattr(config_instance, var.name, dump_default)
                if is_object_type(type(dump_value)):
                    dump_value = {
                        key: (
                            _dump(value, dict_type=dict_type)
                            if is_config(value)
                            else value
                        )
                        for (key, value) in dump_value.items()
                    }

                if dump_value is not None:
                    result[dump_key] = dump_value

    return result


def validate(instance):
    """ Validates a given ``instance``.

    :param object instance: The instance to validate
    :raises jsonschema.exceptions.ValidationError: On failed validation
    """

    jsonschema.validate(
        to_dict(instance, dict_type=dict), build_schema(instance.__class__)
    )


def from_dict(config_cls, dictionary):
    """ Loads an instance of ``config_cls`` from a dictionary.

    :param type config_cls: The class to build an instance of
    :param dict dictionary: The dictionary to load from
    :return: An instance of ``config_cls``
    :rtype: object
    """

    return _build(config_cls, dictionary)


def to_dict(instance, dict_type=OrderedDict):
    """ Dumps an instance to an dictionary mapping.

    :param object instance: The instance to dump
    :param object dict_type: The dictionary type to use, defaults to ``OrderedDict``
    :return: Dictionary serialization of instance
    :rtype: OrderedDict
    """

    return _dump(instance, dict_type=dict_type)



# Copyright (c) 2018 Stephen Bunn <stephen@bunn.io>
# MIT License <https://opensource.org/licenses/MIT>

from ._common import BaseHandler


class PickleHandler(BaseHandler):
    """ The Pickle serialization handler.
    """

    name = "pickle"
    packages = ("pickle",)
    options = {}

    def on_pickle_dumps(self, pickle, dictionary):
        """ The :mod:`pickle` dumps method.

        :param module pickle: The ``pickle`` module
        :param dict dictionary: The dictionary instance to serailize
        :returns: The serialized content
        :rtype: str
        """

        return pickle.dumps(dictionary, protocol=pickle.HIGHEST_PROTOCOL)

    def on_pickle_loads(self, pickle, content):
        """ The :mod:`pickle` loads method.

        :param module pickle: The ``pickle`` module
        :param str content: The content to deserialize
        :returns: The deserialized dictionary
        :rtype: dict
        """

        return pickle.loads(content)



# Copyright (c) 2018 Stephen Bunn <stephen@bunn.io>
# MIT License <https://opensource.org/licenses/MIT>

from .json import JSONHandler
from .toml import TOMLHandler
from .yaml import YAMLHandler
from .pickle import PickleHandler
from .msgpack import MsgpackHandler

__all__ = [
    "JSONHandler",
    "TOMLHandler",
    "YAMLHandler",
    "PickleHandler",
    "MsgpackHandler",
]



# Copyright (c) 2018 Stephen Bunn <stephen@bunn.io>
# MIT License <https://opensource.org/licenses/MIT>

import fnmatch
import warnings

from ._common import BaseHandler


class TOMLHandler(BaseHandler):
    """ The TOML serialization handler.
    """

    name = "toml"
    packages = ("tomlkit", "toml", "pytoml")
    options = {"inline_tables": []}

    def on_tomlkit_dumps(self, tomlkit, dictionary, **kwargs):
        """ The `tomlkit <https://pypi.org/project/tomlkit/>`_ dumps method.

        :param module tomlkit: The ``tomlkit`` module
        :param dict dictionary: The dictionary to serialize
        :param list inline_tables: A list glob patterns to use for derminining which
            dictionaries should be rendered as inline tables, defaults to [], optional
        :returns: The TOML serialization
        :rtype: str

        Dumping inline tables uses :mod:`fnmatch` to compare ``.`` delimited dictionary
        path glob patterns to filter tables

        >>> config.dumps_toml(prefer="tomlkit")
        name = "My Project"
        type = "personal-project"
        keywords = ["example", "test"]
        [dependencies]
        [dependencies.a-dependency]
        name = "A Dependency"
        version = "v12"
        >>> config.dumps_toml(prefer="tomlkit", inline_tables=["dependencies"])
        name = "My Project"
        type = "personal-project"
        keywords = ["example", "test"]
        dependencies = {a-dependency = {name = "A Dependency",version = "v12"}}
        >>> config.dumps_toml(prefer="tomlkit", inline_tables=["dependencies.*"])
        name = "My Project"
        type = "personal-project"
        keywords = ["example", "test"]
        [dependencies]
        a-dependency = {name = "A Dependency",version = "v12"}
        """

        inline_tables = set(kwargs.get("inline_tables", []))

        def _dump_dict(dictionary, source, source_path=[]):
            for (key, value) in dictionary.items():
                if isinstance(value, dict):
                    # checks the current path with fnmatch to see if current table
                    # should be an inline table
                    is_inline = any(
                        [
                            fnmatch.fnmatch(".".join(source_path + [key]), pattern)
                            for pattern in inline_tables
                        ]
                    )
                    if is_inline:
                        table = tomlkit.inline_table()
                        # NOTE: manual dictionary assignment because `tomlkit` does not
                        # impelment `dict.update`
                        for (inline_key, inline_value) in value.items():
                            if isinstance(inline_value, dict):
                                table[inline_key] = _dump_dict(
                                    inline_value,
                                    tomlkit.inline_table(),
                                    source_path=source_path + [inline_key],
                                )
                            else:
                                table[inline_key] = inline_value
                        source[key] = table
                    else:
                        source[key] = _dump_dict(
                            value, tomlkit.table(), source_path=source_path + [key]
                        )
                else:
                    source[key] = value
            return source

        return tomlkit.dumps(_dump_dict(dictionary, tomlkit.document()))

    def on_tomlkit_loads(self, tomlkit, content):
        """ The `tomlkit <https://pypi.org/project/tomlkit/>`_ loads method.

        :param module tomlkit: The ``tomlkit`` module
        :param str content: The content to deserialize
        :returns: The deserialized dictionary
        :rtype: dict
        """

        return tomlkit.parse(content)

    def on_toml_dumps(self, toml, dictionary, **kwargs):
        """ The `toml <https://pypi.org/project/toml/>`_ dumps method.

        :param module toml: The ``toml`` module
        :param dict dictionary: The dictionary to serialize
        :param list inline_tables: A list glob patterns to use for derminining which
            dictionaries should be rendered as inline tables, defaults to [], optional
        :returns: The TOML serialization
        :rtype: str

        Dumping inline tables uses :mod:`fnmatch` to compare ``.`` delimited dictionary
        path glob patterns to filter tables

        >>> config.dumps_toml(prefer="toml")
        name = "My Project"
        type = "personal-project"
        keywords = [ "example", "test",]
        [dependencies.a-dependency]
        name = "A Dependency"
        version = "v12"
        >>> config.dumps_toml(prefer="toml", inline_tables=["dependencies"])
        name = "My Project"
        type = "personal-project"
        keywords = [ "example", "test",]
        dependencies = {a-dependency = {name = "A Dependency",version = "v12"}
        }
        >>> config.dumps_toml(prefer="toml", inline_tables=["dependencies.*"])
        name = "My Project"
        type = "personal-project"
        keywords = [ "example", "test",]
        [dependencies]
        a-dependency = { name = "A Dependency", version = "v12" }
        """

        inline_tables = set(kwargs.get("inline_tables", []))

        def _dump_dict(dictionary, source, source_path=[]):
            for (key, value) in dictionary.items():
                if isinstance(value, dict):
                    is_inline = any(
                        [
                            fnmatch.fnmatch(".".join(source_path + [key]), pattern)
                            for pattern in inline_tables
                        ]
                    )
                    if is_inline:
                        source[key] = toml.TomlDecoder().get_empty_inline_table()
                    else:
                        source[key] = {}
                    source[key].update(
                        _dump_dict(value, {}, source_path=source_path + [key])
                    )
                else:
                    source[key] = value
            return source

        encoder = toml.TomlEncoder(preserve=True)
        return toml.dumps(_dump_dict(dictionary, {}), encoder=encoder)

    def on_toml_loads(self, toml, content):
        """ The `toml <https://pypi.org/project/toml/>`_ loads method.

        :param module toml: The ``toml`` module
        :param str content: The content to deserialize
        :returns: The deserialized dictionary
        :rtype: dict
        """

        return toml.loads(content)

    def on_pytoml_dumps(self, pytoml, dictionary, **kwargs):
        """ The `pytoml <https://pypi.org/project/pytoml/>`_ dumps method.

        :param module pytoml: The ``pytoml`` module
        :param dict dictionary: The dictionary to serialize
        :returns: The TOML serialization
        :rtype: str
        """

        inline_tables = set(kwargs.get("inline_tables", []))
        if len(inline_tables) > 0:
            warnings.warn("pytoml does not support 'inline_tables' argument")
        return pytoml.dumps(dictionary)

    def on_pytoml_loads(self, pytoml, content):
        """ The `pytoml <https://pypi.org/project/pytoml/>`_ loads method.

        :param module pytoml: The ``pytoml`` module
        :param str content: The content to deserialize
        :returns: The deserialized dictionary
        :rtype: dict
        """

        return pytoml.loads(content)



# Copyright (c) 2018 Stephen Bunn (stephen@bunn.io)
# MIT License <https://choosealicense.com/licenses/mit/>

import os
import sys
import codecs
import shutil

import setuptools

VERSION = "0.0.7"

INSTALL_REQUIRES = ["attrs", "jsonschema"]
SETUP_REQUIRES = []
EXTRAS_REQUIRE = {
    "toml": ["toml"],
    "pytoml": ["pytoml"],
    "tomlkit": ["tomlkit"],
    "ujson": ["ujson"],
    "pyyaml": ["pyyaml"],
    "msgpack": ["msgpack"],
}

all_extras = list(set().union(*[packages for packages in EXTRAS_REQUIRE.values()]))
EXTRAS_REQUIRE["docs"] = all_extras + ["sphinx"]
EXTRAS_REQUIRE["test"] = all_extras + [
    "flake8",
    "pytest",
    "pytest-flake8",
    "pytest-sugar",
    "pytest-xdist",
    "hypothesis",
]
EXTRAS_REQUIRE["dev"] = (
    EXTRAS_REQUIRE["docs"] + EXTRAS_REQUIRE["test"] + ["black", "isort"]
)


class UploadCommand(setuptools.Command):

    description = "Build and publish package"
    user_options = []

    @staticmethod
    def status(status):
        print("... {0}".format(status))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status("removing previous builds")
            shutil.rmtree(
                os.path.join(os.path.abspath(os.path.dirname(__file__)), "dist")
            )
        except FileNotFoundError:
            pass

        self.status("building distribution")
        os.system("{0} setup.py sdist".format(sys.executable))

        self.status("uploading distribution")
        os.system("twine upload dist/*")

        self.status("pushing git tags")
        os.system("git tag v{0}".format(VERSION))
        os.system("git push --tags")

        sys.exit()


long_description = ""
with codecs.open("README.rst", encoding="utf-8") as fp:
    long_description = "\n" + fp.read()

setuptools.setup(
    name="file-config",
    version=VERSION,
    description="An attr's based configuration file abstraction",
    long_description=long_description,
    url="https://github.com/stephen-bunn/file-config",
    license="MIT License",
    author="Stephen Bunn",
    author_email="stephen@bunn.io",
    install_requires=INSTALL_REQUIRES,
    setup_requires=SETUP_REQUIRES,
    extras_require=EXTRAS_REQUIRE,
    include_package_data=True,
    packages=setuptools.find_packages(exclude=["tests.*", "tests"]),
    keywords=["config", "file", "toml", "json", "yaml", "msgpack", "pickle"],
    python_requires=">=3.6",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Programming Language :: Python :: 3.6",
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: Developers",
        "Operating System :: POSIX",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
        "Natural Language :: English",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    cmdclass={"upload": UploadCommand},
)



# Copyright (c) 2018 Stephen Bunn <stephen@bunn.io>
# MIT License <https://opensource.org/licenses/MIT>

import re
import typing
import warnings
import collections
from functools import partial

import attr

from .utils import (
    is_bool_type,
    is_null_type,
    is_array_type,
    is_config_var,
    is_regex_type,
    is_union_type,
    is_config_type,
    is_number_type,
    is_object_type,
    is_string_type,
    is_typing_type,
    is_builtin_type,
    is_integer_type,
    is_compiled_pattern,
)
from .constants import CONFIG_KEY, REGEX_TYPE_NAME


def Regex(pattern):
    """ A custom typing type to store regular expressions for schema building.

    :param str pattern: The regular expression
    :return: A new Regex type instance
    :rtype: typing.Type
    """

    return typing.NewType(REGEX_TYPE_NAME, re.compile(pattern))


def _build_attribute_modifiers(
    var, attribute_mapping, ignore=["type", "name", "required", "default"]
):
    if not is_config_var(var):
        raise ValueError(
            f"cannot build field modifiers for {var!r}, is not a config var"
        )

    entry = var.metadata[CONFIG_KEY]
    modifiers = {}

    for (entry_attribute, entry_value) in zip(
        attr.fields(type(entry)), attr.astuple(entry)
    ):
        if entry_value is not None:
            if entry_attribute.name in ignore:
                continue
            elif entry_attribute.name in attribute_mapping:
                # NOTE: required for `isinstance(True, (int, float)) == True`
                if type(entry_value) == entry_attribute.type:
                    modifiers[attribute_mapping[entry_attribute.name]] = entry_value
                else:
                    raise ValueError(
                        f"invalid modifier type for modifier {entry_attribute.name!r} "
                        f"on var {var.name!r}, expected type {entry_attribute.type!r}, "
                        f"received {entry_value!r} of type {type(entry_value)!r}"
                    )
            else:
                warnings.warn(
                    f"field modifier {entry_attribute.name!r} has no effect on var "
                    f"{var.name!r} of type {entry.type!r}"
                )

    return modifiers


def _build_null_type(var, property_path=[]):
    return {"type": "null"}


def _build_bool_type(var, property_path=[]):
    return {"type": "boolean"}


def _build_string_type(var, property_path=[]):
    schema = {"type": "string"}
    if is_builtin_type(var):
        return schema

    if is_regex_type(var):
        schema["pattern"] = var.__supertype__.pattern
        return schema

    schema.update(
        _build_attribute_modifiers(var, {"min": "minLength", "max": "maxLength"})
    )

    if is_regex_type(var.type):
        schema["pattern"] = var.type.__supertype__.pattern
    return schema


def _build_integer_type(var, property_path=[]):
    schema = {"type": "integer"}
    if is_builtin_type(var):
        return schema

    schema.update(_build_attribute_modifiers(var, {"min": "minimum", "max": "maximum"}))
    return schema


def _build_number_type(var, property_path=[]):
    schema = {"type": "number"}
    if is_builtin_type(var):
        return schema

    schema.update(_build_attribute_modifiers(var, {"min": "minimum", "max": "maximum"}))
    return schema


def _build_array_type(var, property_path=[]):
    schema = {"type": "array", "items": {"$id": f"#/{'/'.join(property_path)}/items"}}
    if is_builtin_type(var):
        return schema

    schema.update(
        _build_attribute_modifiers(
            var,
            {
                "min": "minItems",
                "max": "maxItems",
                "unique": "uniqueItems",
                "contains": "contains",
            },
        )
    )

    if is_typing_type(var.type) and len(var.type.__args__) > 0:
        # NOTE: typing.List only allows one typing argument
        nested_type = var.type.__args__[0]
        schema["items"].update(
            _build(nested_type, property_path=property_path + ["items"])
        )
    return schema


def _build_object_type(var, property_path=[]):
    schema = {"type": "object"}
    if is_builtin_type(var):
        return schema

    entry = var.metadata[CONFIG_KEY]

    if isinstance(entry.min, int):
        schema["minProperties"] = entry.min
    if isinstance(entry.max, int):
        schema["maxProperties"] = entry.max

    # NOTE: typing.Dict only accepts two typing arguments
    if is_typing_type(var.type) and len(var.type.__args__) == 2:
        (key_type, value_type) = var.type.__args__

        key_pattern = "^(.*)$"
        if is_regex_type(key_type):
            key_pattern = key_type.__supertype__.pattern
        elif not is_string_type(key_type):
            raise ValueError(
                f"cannot serialize object with key of type {key_type!r}, "
                f"located in var {var.name!r}"
            )

        schema["patternProperties"] = {
            key_pattern: _build(value_type, property_path=property_path)
        }

    return schema


def _build_type(type_, value, property_path=[]):
    for (type_check, builder) in (
        (is_null_type, _build_null_type),
        (is_bool_type, _build_bool_type),
        (is_string_type, _build_string_type),
        (is_integer_type, _build_integer_type),
        (is_number_type, _build_number_type),
        (is_array_type, _build_array_type),
        (is_object_type, _build_object_type),
    ):
        if type_check(type_):
            return builder(value, property_path=property_path)

    # NOTE: warning ignores type None (as that is the config var default)
    if type_:
        warnings.warn(f"unhandled translation for type {type_!r} with value {value!r}")
    return {}


def _build_var(var, property_path=[]):
    if not is_config_var(var):
        raise ValueError(f"var {var!r} is not a config var")

    entry = var.metadata[CONFIG_KEY]
    var_name = entry.name if entry.name else var.name
    schema = {"$id": f"#/{'/'.join(property_path)}/{var_name}"}

    if var.default is not None:
        schema["default"] = var.default

    if entry is not None:
        if isinstance(entry.title, str):
            schema["title"] = entry.title
        if isinstance(entry.description, str):
            schema["description"] = entry.description
        if isinstance(entry.examples, collections.Iterable) and len(entry.examples) > 0:
            schema["examples"] = entry.examples

    # handle typing.Union types by simply using the "anyOf" key
    if is_union_type(var.type):
        type_union = {"anyOf": []}
        for allowed_type in var.type.__args__:
            type_union["anyOf"].append(
                _build_type(
                    allowed_type, allowed_type, property_path=property_path + [var_name]
                )
            )
        schema.update(type_union)
    else:
        schema.update(
            _build_type(var.type, var, property_path=property_path + [var_name])
        )
    return schema


def _build_config(config_cls, property_path=[]):
    if not is_config_type(config_cls):
        raise ValueError(f"class {config_cls!r} is not a config class")

    schema = {"type": "object", "required": [], "properties": {}}
    cls_entry = getattr(config_cls, CONFIG_KEY)

    # add schema title, defaults to config classes `__qualname__`
    schema_title = cls_entry.get("title", config_cls.__qualname__)
    if isinstance(schema_title, str):
        schema["title"] = schema_title

    schema_description = cls_entry.get("description")
    if isinstance(schema_description, str):
        schema["description"] = schema_description

    # if the length of the property path is 0, assume that current object is root
    if len(property_path) <= 0:
        schema["$id"] = f"{config_cls.__qualname__}.json"
        schema["$schema"] = "http://json-schema.org/draft-07/schema#"
    else:
        schema["$id"] = f"#/{'/'.join(property_path)}"

    property_path.append("properties")
    for var in attr.fields(config_cls):
        if not is_config_var(var):
            # encountered attribute is not a serialized field (i.e. missing CONFIG_KEY)
            continue
        entry = var.metadata[CONFIG_KEY]
        var_name = entry.name if entry.name else var.name
        if entry.required:
            schema["required"].append(var_name)

        if is_config_type(var.type):
            schema["properties"][var_name] = _build_config(
                var.type, property_path=property_path + [var_name]
            )
        else:
            schema["properties"][var_name] = _build_var(
                var, property_path=property_path
            )

    return schema


def _build(value, property_path=[]):
    if is_config_type(value):
        return _build_config(value, property_path=property_path)
    elif is_config_var(value):
        return _build_var(value, property_path=property_path)
    elif is_builtin_type(value):
        return _build_type(value, value, property_path=property_path)
    elif is_regex_type(value):
        # NOTE: building regular expression types assumes type is string
        return _build_type(str, value, property_path=property_path)
    return _build_type(type(value), value, property_path=property_path)


def build_schema(config_cls):
    """ Builds the JSONSchema for a given config class.

    :param class config_cls: The config class to build the JSONSchema for
    :return: The resulting JSONSchema
    :rtype: dict
    """

    return _build_config(config_cls, property_path=[])



# Copyright (c) 2018 Stephen Bunn <stephen@bunn.io>
# MIT License <https://opensource.org/licenses/MIT>

from . import handlers, __version__
from .constants import CONFIG_KEY
from ._file_config import var, config, to_dict, validate, from_dict, make_config
from .schema_builder import Regex, build_schema
